"""Developer API.

.. autosummary::
   :toctree:

   UserSettings
   InstanceSettings
   Storage

"""
from ._deprecated import deprecated
from ._docs import doc_args
from ._settings_instance import InstanceSettings
from ._settings_user import UserSettings
from ._storage import Storage
from ._testdb import setup_local_test_postgres, setup_local_test_sqlite_file
from ._upath_ext import UPath



"""LaminDB manager.

The Python API here has a 1:1 correspondnce to the CLI tool `lndb`.

Import the package::

   import lndb

Settings:

.. autosummary::
   :toctree:

   settings

Setup user account:

.. autosummary::
   :toctree:

   signup
   login

Setup instance:

.. autosummary::
   :toctree:

   init
   load

More instance operations:

.. autosummary::
   :toctree:

   close
   delete
   info
   set

Manage creating and testing migrations (deployment is automatic):

.. autosummary::
   :toctree:

   migrate


Developer API.

.. autosummary::
   :toctree:

   dev
"""

__version__ = "0.35rc2"  # denote a pre-release for 0.1.0 with 0.1rc1

import sys
from os import name as _os_name

from . import _check_versions  # noqa
from . import dev
from ._close import close  # noqa
from ._delete import delete  # noqa
from ._info import info  # noqa
from ._init_instance import init  # noqa
from ._load_instance import load  # noqa
from ._migrate import migrate
from ._schema import schema  # noqa
from ._set import set, set_storage  # noqa
from ._settings import settings  # noqa
from ._setup_user import login, signup  # noqa


# unlock and clear even if an uncaught exception happens
def _clear_on_exception(typ, value, traceback):
    from .dev._exclusion import _locker

    if _locker is not None:
        try:
            _locker._clear()
        except:
            pass
    sys.__excepthook__(typ, value, traceback)


sys.excepthook = _clear_on_exception

# hide the supabase error in a thread on windows
if _os_name == "nt":
    if sys.version_info.minor > 7:
        import threading

        _original_excepthook = threading.excepthook

        def _except_hook(args):
            is_overflow = args.exc_type is OverflowError
            for_timeout = str(args.exc_value) == "timeout value is too large"
            if not (is_overflow and for_timeout):
                _original_excepthook(args)

        threading.excepthook = _except_hook



"""Check and auto-deploy migrations."""

import importlib
import os
from pathlib import Path
from subprocess import run
from typing import Any, Optional

import sqlmodel as sqm
from lamin_logger import logger
from natsort import natsorted
from packaging.version import parse as vparse

from ..dev._db import insert
from ..dev._settings_instance import InstanceSettings
from ..dev._settings_user import UserSettings
from ..dev._setup_schema import create_schema_if_not_exists, get_schema_module_name
from ._utils import generate_module_files, modify_alembic_ini


def check_deploy_migration(
    *,
    usettings: UserSettings,
    isettings: InstanceSettings,
    migrate_confirmed: Optional[bool] = None,
):
    if "LAMIN_SKIP_MIGRATION" in os.environ:
        if os.environ["LAMIN_SKIP_MIGRATION"] == "true":
            return "migrate-skipped"

    # lock the whole migration
    locker = isettings._cloud_sqlite_locker
    locker.lock()
    # synchronize the sqlite file before proceeding
    isettings._update_local_sqlite_file()

    status = []
    schema_names = ["core"] + list(isettings.schema)

    for schema_name in schema_names:
        create_schema_if_not_exists(schema_name, isettings)
        schema_module_name = get_schema_module_name(schema_name)
        schema_module = importlib.import_module(schema_module_name)
        if schema_module._migration is None:
            status.append("migrate-unnecessary")
            continue

        schema_id = schema_module._schema_id

        with sqm.Session(isettings.engine) as session:
            table_loc = (
                schema_module.dev if hasattr(schema_module, "dev") else schema_module
            )
            version_table = getattr(table_loc, f"version_{schema_id}")
            versions = session.exec(sqm.select(version_table.v)).all()
            versions = natsorted(versions)  # latest version is last

        current_version = schema_module.__version__

        if current_version not in versions and len(versions) > 0:
            if vparse(current_version) < vparse(versions[-1]):  # type: ignore
                raise RuntimeError(
                    "You are trying to connect to a DB that already runs"
                    f" v{versions[-1]} of {schema_module_name} but you only have"
                    f" v{current_version} installed.\nPlease run `pip install"
                    f" {schema_module_name}=={versions[-1]}`, or install the latest"
                    " schema module version from GitHub."
                )
            logger.info(
                f"Schema {schema_name} v{versions[-1]} is not up to date"
                f" with {current_version}."
            )
            # if migration is confirmed, continue
            if migrate_confirmed:
                pass
            # run a confirmation dialogue outside a pytest run
            elif "PYTEST_CURRENT_TEST" not in os.environ:
                logger.warning(
                    "Run the command in the shell to respond to the following dialogue."
                )

                response = input(
                    f"Do you want to migrate {schema_name} from {versions[-1]} to"
                    f" {current_version} (y/n)?"
                )

                if response != "y":
                    logger.warning(
                        f"Your db does not match the latest version of schema {schema_name}.\n"  # noqa
                        "For production use, either install"
                        f" {schema_module_name} {versions[-1]} "
                        f"or migrate the database to {current_version}."
                    )
                    return None
            else:
                logger.warning(
                    f"Migrate instance {isettings.name} outside a test (CI) run. "
                    "Unexpected errors might happen."
                )
                return "migrate-failed"

            generate_module_files(
                package_name=schema_module_name,
                migrations_path=Path(schema_module.__file__).parent / "migrations",  # type: ignore  # noqa
                schema_id=schema_id,
            )
            migrate_status = deploy(
                version=current_version,
                usettings=usettings,
                isettings=isettings,
                schema_name=schema_name,
                schema_id=schema_id,
                schema_module=schema_module,
            )
            status.append(migrate_status)
        else:
            status.append("migrate-unnecessary")

    locker.unlock()

    if "migrate-failed" in status:
        return "migrate-failed"
    elif "migrate-success" in status:
        return "migrate-success"
    else:
        return "migrate-unnecessary"


def deploy(
    *,
    version: str,
    usettings: UserSettings,
    isettings: InstanceSettings,
    schema_name: str,
    schema_id: str,
    schema_module: Any,
):
    """Migrate database to latest version."""
    schema_root = Path(schema_module.__file__).parent
    filepath = schema_root / "alembic.ini"

    modify_alembic_ini(filepath, isettings, schema_name)

    process = run(
        f"python -m alembic --name {schema_id} upgrade head",
        cwd=f"{schema_root}",
        shell=True,
    )

    if process.returncode == 0:
        logger.success(f"Successfully migrated schema {schema_name} to v{version}.")
        # The following call will also update the sqlite file in the cloud.
        insert.version(
            schema_module=schema_module,
            user_id=usettings.id,  # type: ignore
        )
    else:
        logger.error("Automatic migration failed.")

    modify_alembic_ini(filepath, isettings, schema_name, revert=True)

    if process.returncode == 0:
        return "migrate-success"
    else:
        return "migrate-failed"



import importlib
from subprocess import run
from typing import Optional

from lamin_logger import logger
from packaging import version

from lndb._assets import instances as test_instances
from lndb._init_instance import init
from lndb._settings import settings
from lndb.dev import setup_local_test_sqlite_file
from lndb.dev._clone import clone_test
from lndb.dev._settings_instance import InstanceSettings


def migrate_clones(
    schema_package: str, n_instances: Optional[int] = None, dialect_name="sqlite"
):
    # auto-bump version to simulate state after release
    schema_module = importlib.import_module(schema_package)
    v = version.parse(schema_module.__version__)
    schema_module.__version__ = f"{v.major}.{v.minor+1}.0"  # type: ignore  # noqa
    # get test instances
    instances = test_instances.loc[test_instances[1] == schema_package][0]
    if dialect_name == "sqlite":
        run_instances = instances.loc[
            instances.str.startswith("s3://") | instances.str.startswith("gc://")
        ]
    elif dialect_name == "postgresql":
        run_instances = instances.loc[instances.str.startswith(dialect_name)]
    else:
        raise ValueError("Pass either 'sqlite' or 'postgresql'.")
    display_list = [inst.split("/")[-1] for inst in run_instances.tolist()]
    logger.info(f"These instances need to be tested: {display_list}")
    results = []
    for instance in run_instances.iloc[:n_instances]:
        logger.info(f"Testing: {instance}")
        db, storage_root, name = None, None, None
        for prefix in {"s3://", "gc://"}:
            if instance.startswith(prefix):
                db = None
                name = instance.replace(prefix, "").split("/")[0]
                storage_root = prefix + name
        if instance.startswith("postgresql"):
            db = instance
            storage_root = "pgtest"
            name = "pgtest"
        # init test instance
        src_settings = InstanceSettings(
            storage_root=storage_root,
            db=db,
            name=name,  # type: ignore  # noqa
            owner=settings.user.handle,
        )
        connection_string = clone_test(src_settings=src_settings)
        if db is None:
            storage_test = setup_local_test_sqlite_file(src_settings, return_dir=True)
            result = init(storage=storage_test, _migrate=True)
        else:
            result = init(db=connection_string, storage=storage_root, _migrate=True)
        logger.info(result)
        if dialect_name == "postgresql":
            run("docker stop pgtest && docker rm pgtest", shell=True)
        results.append(result)
    return results



from pathlib import Path
from typing import Any, Dict, get_type_hints

from ._settings_store import (
    UserSettingsStore,
    current_user_settings_file,
    user_settings_file_email,
    user_settings_file_handle,
)
from ._settings_user import UserSettings


def save_user_settings(settings: UserSettings):
    assert settings.email is not None
    type_hints = get_type_hints(UserSettingsStore)
    save_settings(settings, current_user_settings_file(), type_hints)
    save_settings(settings, user_settings_file_email(settings.email), type_hints)
    if settings.handle is not None:
        save_settings(
            settings, user_settings_file_handle(settings.handle), type_hints
        )  # noqa


def save_settings(
    settings: Any,
    settings_file: Path,
    type_hints: Dict[str, Any],
):
    with open(settings_file, "w") as f:
        for store_key, type in type_hints.items():
            if "__" not in store_key:
                if store_key in {"dbconfig_", "schema_", "name_"}:
                    settings_key = f"_{store_key.rstrip('_')}"
                else:
                    settings_key = store_key
                value = getattr(settings, settings_key)
                if value is None:
                    value = "null"
                else:
                    value = type(value)
                f.write(f"{store_key}={value}\n")


def save_instance_settings(settings: Any, settings_file: Path):
    with open(settings_file, "w") as f:
        f.write(f"owner={settings.owner}\n")
        f.write(f"name={settings.name}\n")
        f.write(f"storage_root={str(settings.storage.root)}\n")
        storage_region = (
            settings.storage.region if settings.storage.region is not None else "null"
        )
        f.write(f"storage_region={storage_region}\n")
        db = settings._db if settings._db is not None else "null"
        f.write(f"db={db}\n")
        schema_str = (
            settings._schema_str if settings._schema_str is not None else "null"
        )
        f.write(f"schema_str={schema_str}\n")



from lnhub_rest import __version__ as lnhub_rest_v
from packaging import version

if version.parse(lnhub_rest_v) < version.parse("0.4.3"):
    raise SystemExit("Please upgrade lnhub_rest: pip install lnhub_rest -U")



from pathlib import Path
from typing import Literal, Optional, Union

from appdirs import AppDirs

from ._upath_ext import UPath

DIRS = AppDirs("lamindb", "laminlabs")


class Storage:
    """Manage cloud or local storage."""

    def __init__(self, root: Union[str, Path, UPath], region: Optional[str] = None):
        if isinstance(root, str):
            root_path = Storage._str_to_path(root)
        else:
            root_path = root
        self._root = root_path
        self._region = region

    @staticmethod
    def _str_to_path(storage: str) -> Union[Path, UPath]:
        if storage.startswith("s3://"):
            # for new buckets there could be problems if the region is not specified
            storage_root = UPath(storage, cache_regions=True)
        elif storage.startswith("gs://"):
            storage_root = UPath(storage)
        else:  # local path
            storage_root = Path(storage).absolute()
        return storage_root

    @property
    def root(self) -> Union[Path, UPath]:
        """Root storage location."""
        return self._root

    @property
    def cache_dir(
        self,
    ) -> Union[Path, None]:
        """Cache root, a local directory to cache cloud files."""
        if self.is_cloud:
            cache_dir = Path(DIRS.user_cache_dir)
            cache_dir.mkdir(parents=True, exist_ok=True)
        else:
            cache_dir = None
        return cache_dir

    @property
    def is_cloud(self) -> bool:
        """`True` if `storage_root` is in cloud, `False` otherwise."""
        return isinstance(self.root, UPath)

    @property
    def region(self) -> Optional[str]:
        """Storage region."""
        return self._region

    @property
    def type(self) -> Literal["s3", "gs", "local"]:
        """AWS S3 vs. Google Cloud vs. local.

        Returns "s3" or "gs" or "local".
        """
        root_str = str(self.root)
        if root_str.startswith("s3://"):
            return "s3"
        elif root_str.startswith("gs://"):
            return "gs"
        else:
            return "local"

    def key_to_filepath(self, filekey: Union[Path, UPath, str]) -> Union[Path, UPath]:
        """Cloud or local filepath from filekey."""
        return self.root / filekey

    def cloud_to_local(self, filepath: Union[Path, UPath]) -> Path:
        """Local (cache) filepath from filepath."""
        local_filepath = self.cloud_to_local_no_update(filepath)  # type: ignore
        if isinstance(filepath, UPath):
            local_filepath.parent.mkdir(parents=True, exist_ok=True)
            filepath.synchronize(local_filepath)
        return local_filepath

    # conversion to Path via cloud_to_local() would trigger download
    # of remote file to cache if there already is one
    # in pure write operations that update the cloud, we don't want this
    # hence, we manually construct the local file path
    # using the `.parts` attribute in the following line
    def cloud_to_local_no_update(self, filepath: Union[Path, UPath]) -> Path:
        if self.is_cloud:
            return self.cache_dir.joinpath(filepath._url.netloc, *filepath.parts[1:])  # type: ignore # noqa
        return filepath

    def local_filepath(self, filekey: Union[Path, UPath, str]) -> Path:
        """Local (cache) filepath from filekey: `local(filepath(...))`."""
        return self.cloud_to_local(self.key_to_filepath(filekey))



import shutil
from pathlib import Path

from lamin_logger import logger

from ._settings import settings
from .dev._settings_load import load_instance_settings
from .dev._settings_store import instance_settings_file


def delete(instance_name: str):
    """Delete an instance."""
    logger.info(f"Deleting instance {settings.user.handle}/{instance_name}")
    settings_file = instance_settings_file(instance_name, settings.user.handle)
    isettings = load_instance_settings(settings_file)

    delete_settings(settings_file)
    delete_cache(isettings.storage.cache_dir)
    logger.info(
        f"    consider deleting your stored data manually: {isettings.storage.root}"
    )
    if isettings.is_remote:
        logger.info("    please manually delete your remote instance on lamin.ai")
    else:
        if isettings.dialect == "sqlite":
            isettings._sqlite_file.unlink()
            logger.info("    deleted sqlite file")


def delete_cache(cache_dir: Path):
    if cache_dir is not None and cache_dir.exists():
        shutil.rmtree(cache_dir)
        logger.info("    instance cache deleted")


def delete_settings(settings_file: Path):
    settings_file.unlink()
    logger.info("    instance settings deleted")



import os
from typing import Optional

from lamin_logger import logger
from lnhub_rest._init_instance import init_instance as init_instance_hub
from lnhub_rest._load_instance import load_instance as load_instance_from_hub

from ._settings import InstanceSettings, settings
from .dev._settings_load import load_instance_settings
from .dev._settings_store import instance_settings_file


def load(
    identifier: str,
    *,
    migrate: Optional[bool] = None,
    _log_error_message: bool = True,
) -> Optional[str]:
    """Load existing instance.

    Args:
        identifier: The instance identifier can the instance name (owner is
            current user), handle/name, or the URL: https://lamin.ai/handle/name.
        migrate: Whether to auto-migrate or not.

    Returns:
        - migrate-failed if migration failed
        - migrate-success if migration was successful
        - migrate-unnecessary if migration was not required
    """
    owner, name = get_owner_name_from_identifier(identifier)

    hub_result = load_instance_from_hub(owner=owner, name=name)
    if not isinstance(hub_result, str):
        instance, storage = hub_result
        isettings = InstanceSettings(
            owner=owner,
            name=name,
            storage_root=storage.root,
            storage_region=storage.region,
            db=instance.db,
            schema=instance.schema_str,
        )
    else:
        settings_file = instance_settings_file(name, owner)
        if settings_file.exists():
            isettings = load_instance_settings(settings_file)
        else:
            if _log_error_message:
                logger.error("Instance neither exists on hub nor locally, call `init`.")
            return "instance-not-exists"

    check, msg = isettings._is_db_setup()
    if not check:
        if _log_error_message:
            raise RuntimeError(msg)
        else:
            logger.warning(
                "Instance metadata exists, but DB might have been corrupted or deleted."
                " Re-initializing the DB."
            )
            return "instance-not-exists"

    # register legacy instances on hub if they aren't yet!
    if isettings.is_remote and isinstance(hub_result, str):
        logger.info("Registering instance on hub.")
        init_instance_hub(
            owner=isettings.owner,
            name=isettings.name,
            storage=str(isettings.storage.root),
            db=isettings._db,
            schema=isettings._schema_str,
        )

    message = load_from_isettings(isettings, migrate)

    if not message == "migrate-failed":
        os.environ["LAMINDB_INSTANCE_LOADED"] = "1"

    return message


def get_owner_name_from_identifier(identifier: str):
    if "/" in identifier:
        if identifier.startswith("https://lamin.ai/"):
            identifier = identifier.replace("https://lamin.ai/", "")
        split = identifier.split("/")
        if len(split) > 2:
            raise ValueError(
                "The instance identifier needs to be 'owner/name', the instance name"
                " (owner is current user) or the URL: https://lamin.ai/owner/name."
            )
        owner, name = split
    else:
        owner = settings.user.handle
        name = identifier
    return owner, name


def load_from_isettings(
    isettings: InstanceSettings,
    migrate: Optional[bool] = None,
):
    from ._init_instance import persist_check_reload_schema, register
    from ._migrate import check_deploy_migration
    from .dev._setup_knowledge import load_bionty_versions

    logger.info(f"Loading instance: {isettings.owner}/{isettings.name}")
    persist_check_reload_schema(isettings)
    message = check_deploy_migration(
        usettings=settings.user, isettings=isettings, migrate_confirmed=migrate
    )
    if message == "migrate-failed":
        return message
    register(isettings, settings.user)
    load_bionty_versions(isettings)
    return message



import os
from typing import Union

from .dev._settings_instance import InstanceSettings
from .dev._settings_load import load_instance_settings, load_or_create_user_settings
from .dev._settings_user import UserSettings


# https://stackoverflow.com/questions/128573/using-property-on-classmethods/64738850#64738850
class classproperty(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        """Get."""
        return self.fget(owner_cls)


class settings:
    """Settings access.

    - :class:`~lndb.dev.InstanceSettings`
    - :class:`~lndb.dev.UserSettings`
    """

    _user_settings: Union[UserSettings, None] = None
    _instance_settings: Union[InstanceSettings, None] = None

    _user_settings_env: Union[str, None] = None
    _instance_settings_env: Union[str, None] = None

    @classproperty
    def user(cls) -> UserSettings:
        """User-related settings."""
        if (
            cls._user_settings is None
            or cls._user_settings_env != get_env_name()  # noqa
        ):
            cls._user_settings = load_or_create_user_settings()
            cls._user_settings_env = get_env_name()
            if cls._user_settings and cls._user_settings.id is None:
                raise RuntimeError("Need to login, first: lndb login <email>.")
        return cls._user_settings  # type: ignore

    @classproperty
    def instance(cls) -> InstanceSettings:
        """Instance-related settings."""
        if (
            cls._instance_settings is None
            or cls._instance_settings_env != get_env_name()  # noqa
        ):
            cls._instance_settings = load_instance_settings()
            cls._instance_settings_env = get_env_name()
        return cls._instance_settings  # type: ignore

    @classproperty
    def _instance_exists(cls):
        try:
            cls.instance
            return True
        except RuntimeError:
            return False


def get_env_name():
    if "LAMIN_ENV" in os.environ:
        return os.environ["LAMIN_ENV"]
    else:
        return "prod"



import time
from pathlib import Path
from subprocess import run

from lamin_logger import logger

from ._settings_instance import InstanceSettings


def setup_local_test_sqlite_file(
    src_settings: InstanceSettings, return_dir: bool = False
):
    path = src_settings._sqlite_file_local
    new_stem = path.stem + "_test"
    tgt_sqlite_dir = Path.cwd() / new_stem
    if return_dir:
        return tgt_sqlite_dir
    tgt_sqlite_file = tgt_sqlite_dir / f"{new_stem}{path.suffix}"
    tgt_sqlite_file.parent.mkdir(exist_ok=True)
    if tgt_sqlite_file.exists():
        tgt_sqlite_file.unlink()
    tgt_db = f"sqlite:///{tgt_sqlite_file}"
    return tgt_db


def setup_local_test_postgres(name: str = "pgtest"):
    process = run(
        f"docker run --name {name} -e POSTGRES_PASSWORD=pwd"
        f" -e POSTGRES_DB={name} -d -p 5432:5432 postgres",  # noqa
        shell=True,
    )
    if process.returncode == 0:
        logger.info(
            f"Created Postgres test instance. It runs in docker container '{name}'."
        )
    else:
        raise RuntimeError("Failed to set up postgres test instance.")
    time.sleep(2)
    return f"postgresql://postgres:pwd@0.0.0.0:5432/{name}"


def setup_local_test_postgres_supabase():
    process = run(
        f"supabase start",  # noqa
        shell=True,
    )
    if process.returncode == 0:
        logger.info("Created Supabase test instance.")
    else:
        raise RuntimeError("Failed to set up Supabase test instance.")
    time.sleep(2)
    return "postgresql://postgres:postgres@localhost:54322/postgres"



"""Testing tools for Lamin's Python packages."""
from lndb.test._env import get_package_name
from lndb.test._migrations_e2e import migrate_clones
from lndb.test._migrations_unit import (
    migration_id_is_consistent,
    model_definitions_match_ddl,
)



from ._settings import settings


def info():
    """Log information about current instance & user."""
    # Accessing cached settings is faster than accessing the hub
    print(settings.user)
    print(settings.instance)
